@@ -102,6 +102,11 @@ struct media_player {
 	uint32_t		duration;
 	uint8_t			volume;
 	GTimer			*timer;
+	bool			play;
+	bool			pause;
+	bool			next;
+	bool			previous;
+	bool			control;
 };
 
 static GSList *adapters = NULL;
@@ -127,6 +132,9 @@ static void media_endpoint_cancel(struct endpoint_request *request)
 
 	endpoint->requests = g_slist_remove(endpoint->requests, request);
 
+	if (request->cb)
+		request->cb(endpoint, NULL, -1, request->user_data);
+
 	endpoint_request_free(request);
 }
 
@@ -170,7 +178,7 @@ static struct media_endpoint *media_adapter_find_endpoint(
 		if (path && g_strcmp0(endpoint->path, path) != 0)
 			continue;
 
-		if (uuid && g_strcmp0(endpoint->uuid, uuid) != 0)
+		if (uuid && strcasecmp(endpoint->uuid, uuid) != 0)
 			continue;
 
 		return endpoint;
@@ -368,7 +376,7 @@ static gboolean select_configuration(struct media_endpoint *endpoint,
 	return media_endpoint_async_call(msg, endpoint, cb, user_data, destroy);
 }
 
-static gint transport_device_cmp(gconstpointer data, gconstpointer user_data)
+static int transport_device_cmp(gconstpointer data, gconstpointer user_data)
 {
 	struct media_transport *transport = (struct media_transport *) data;
 	const struct audio_device *device = user_data;
@@ -411,8 +419,8 @@ static gboolean set_configuration(struct media_endpoint *endpoint,
 	if (transport != NULL)
 		return FALSE;
 
-	transport = media_transport_create(endpoint, device,
-						configuration, size);
+	transport = media_transport_create(device, configuration, size,
+								endpoint);
 	if (transport == NULL)
 		return FALSE;
 
@@ -607,13 +615,28 @@ static gboolean endpoint_init_a2dp_sink(struct media_endpoint *endpoint,
 	return TRUE;
 }
 
+static struct media_adapter *find_adapter(struct btd_device *device)
+{
+	GSList *l;
+
+	for (l = adapters; l; l = l->next) {
+		struct media_adapter *adapter = l->data;
+
+		if (adapter->btd_adapter == device_get_adapter(device))
+			return adapter;
+	}
+
+	return NULL;
+}
+
 static bool endpoint_properties_exists(const char *uuid,
 						struct btd_device *dev,
 						void *user_data)
 {
-	struct media_adapter *adapter = user_data;
+	struct media_adapter *adapter;
 
-	if (adapter->btd_adapter != device_get_adapter(dev))
+	adapter = find_adapter(dev);
+	if (adapter == NULL)
 		return false;
 
 	if (media_adapter_find_endpoint(adapter, NULL, NULL, uuid) == NULL)
@@ -659,10 +682,14 @@ static bool endpoint_properties_get(const char *uuid,
 						DBusMessageIter *iter,
 						void *user_data)
 {
-	struct media_adapter *adapter = user_data;
+	struct media_adapter *adapter;
 	DBusMessageIter dict;
 	GSList *l;
 
+	adapter = find_adapter(dev);
+	if (adapter == NULL)
+		return false;
+
 	dbus_message_iter_open_container(iter, DBUS_TYPE_ARRAY,
 					DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
 					DBUS_TYPE_STRING_AS_STRING
@@ -673,7 +700,7 @@ static bool endpoint_properties_get(const char *uuid,
 	for (l = adapter->endpoints; l; l = l->next) {
 		struct media_endpoint *endpoint = l->data;
 
-		if (g_strcmp0(endpoint->uuid, uuid) != 0)
+		if (strcasecmp(endpoint->uuid, uuid) != 0)
 			continue;
 
 		append_endpoint(endpoint, &dict);
@@ -743,7 +770,7 @@ static struct media_endpoint *media_endpoint_create(struct media_adapter *adapte
 		btd_profile_add_custom_prop(uuid, "a{sv}", "MediaEndpoints",
 						endpoint_properties_exists,
 						endpoint_properties_get,
-						adapter);
+						NULL);
 	}
 
 	adapter->endpoints = g_slist_append(adapter->endpoints, endpoint);
@@ -1107,6 +1134,82 @@ static void set_volume(uint8_t volume, struct audio_device *dev, void *user_data
 	}
 }
 
+static bool media_player_send(struct media_player *mp, const char *name)
+{
+	DBusMessage *msg;
+
+	msg = dbus_message_new_method_call(mp->sender, mp->path,
+					MEDIA_PLAYER_INTERFACE, name);
+	if (msg == NULL) {
+		error("Couldn't allocate D-Bus message");
+		return false;
+	}
+
+	g_dbus_send_message(btd_get_dbus_connection(), msg);
+
+	return true;
+}
+
+static bool play(void *user_data)
+{
+	struct media_player *mp = user_data;
+
+	DBG("");
+
+	if (!mp->play || !mp->control)
+		return false;
+
+	return media_player_send(mp, "Play");
+}
+
+static bool stop(void *user_data)
+{
+	struct media_player *mp = user_data;
+
+	DBG("");
+
+	if (!mp->control)
+		return false;
+
+	return media_player_send(mp, "Stop");
+}
+
+static bool pause(void *user_data)
+{
+	struct media_player *mp = user_data;
+
+	DBG("");
+
+	if (!mp->pause || !mp->control)
+		return false;
+
+	return media_player_send(mp, "Pause");
+}
+
+static bool next(void *user_data)
+{
+	struct media_player *mp = user_data;
+
+	DBG("");
+
+	if (!mp->next || !mp->control)
+		return false;
+
+	return media_player_send(mp, "Next");
+}
+
+static bool previous(void *user_data)
+{
+	struct media_player *mp = user_data;
+
+	DBG("");
+
+	if (!mp->previous || !mp->control)
+		return false;
+
+	return media_player_send(mp, "Previous");
+}
+
 static struct avrcp_player_cb player_cb = {
 	.list_settings = list_settings,
 	.get_setting = get_setting,
@@ -1117,7 +1220,12 @@ static struct avrcp_player_cb player_cb = {
 	.get_position = get_position,
 	.get_duration = get_duration,
 	.get_status = get_status,
-	.set_volume = set_volume
+	.set_volume = set_volume,
+	.play = play,
+	.stop = stop,
+	.pause = pause,
+	.next = next,
+	.previous = previous,
 };
 
 static void media_player_exit(DBusConnection *connection, void *user_data)
@@ -1234,8 +1342,12 @@ static gboolean parse_int64_metadata(struct media_player *mp, const char *key,
 {
 	uint64_t value;
 	char valstr[20];
+	int type;
 
-	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_INT64)
+	type = dbus_message_iter_get_arg_type(iter);
+	if (type == DBUS_TYPE_UINT64)
+		warn("expected DBUS_TYPE_INT64 got DBUS_TYPE_UINT64");
+	else if (type != DBUS_TYPE_INT64)
 		return FALSE;
 
 	dbus_message_iter_get_basic(iter, &value);
@@ -1257,8 +1369,12 @@ static gboolean parse_int32_metadata(struct media_player *mp, const char *key,
 {
 	uint32_t value;
 	char valstr[20];
+	int type;
 
-	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_INT32)
+	type = dbus_message_iter_get_arg_type(iter);
+	if (type == DBUS_TYPE_UINT32)
+		warn("expected DBUS_TYPE_INT32 got DBUS_TYPE_UINT32");
+	else if (type != DBUS_TYPE_INT32)
 		return FALSE;
 
 	dbus_message_iter_get_basic(iter, &value);
@@ -1394,7 +1510,7 @@ static const char *loop_status_to_repeat(const char *value)
 		return "off";
 	else if (strcasecmp(value, "Track") == 0)
 		return "singletrack";
-	else if (strcasecmp(value, "Track") == 0)
+	else if (strcasecmp(value, "Playlist") == 0)
 		return "alltracks";
 
 	return NULL;
@@ -1416,6 +1532,21 @@ static gboolean set_repeat(struct media_player *mp, DBusMessageIter *iter)
 	return set_property(mp, "Repeat", value);
 }
 
+static gboolean set_flag(struct media_player *mp, DBusMessageIter *iter,
+								bool *var)
+{
+	dbus_bool_t value;
+
+	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_BOOLEAN)
+		return FALSE;
+
+	dbus_message_iter_get_basic(iter, &value);
+
+	*var = value;
+
+	return TRUE;
+}
+
 static gboolean set_player_property(struct media_player *mp, const char *key,
 							DBusMessageIter *entry)
 {
@@ -1441,6 +1572,21 @@ static gboolean set_player_property(struct media_player *mp, const char *key,
 	if (strcasecmp(key, "LoopStatus") == 0)
 		return set_repeat(mp, &var);
 
+	if (strcasecmp(key, "CanPlay") == 0)
+		return set_flag(mp, &var, &mp->play);
+
+	if (strcasecmp(key, "CanPause") == 0)
+		return set_flag(mp, &var, &mp->pause);
+
+	if (strcasecmp(key, "CanGoNext") == 0)
+		return set_flag(mp, &var, &mp->next);
+
+	if (strcasecmp(key, "CanGoPrevious") == 0)
+		return set_flag(mp, &var, &mp->previous);
+
+	if (strcasecmp(key, "CanControl") == 0)
+		return set_flag(mp, &var, &mp->control);
+
 	DBG("%s not supported, ignoring", key);
 
 	return TRUE;
